/**
Copyright (C) 2003, 2004, 2007 Dominik Wagenfuehr <dominik.wagenfuehr@arcor.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/* Das Programm erstellt aus einem Assembler-Programm v.txt ein
Struktogramm, soweit mglich. Die Ausgabe erfolgt in einer
HTML-Datei, die nachtrglich verndert werden und ausgedruckt
werden kann. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define MID (14)     // MID >= 3,      Anzahl der Spalten fr Verzweigungen
#define COLS (2*MID)                   // Gesamtanzahl der Spalten
#define SPRUNG (0)                     // Spalte fr Sprungmarken
#define REMARKS (COLS-1)               // Spalte fr Bemerkungen
#define TRUE (1)                       // Boolsche Variable wahr
#define FALSE (0)                      // Boolsche Variable falsch

/* Die folgenden Makros definieren die verschiedenen Verzweigungen
und Hintergrnde fr Befehle. Die Abkrzungen sind identisch mit
den Gifs aus dem pics-Ordner. */

#define EM (0)
#define BL (1)
#define QU (2)
#define CL (3)
#define HC (4)
#define VC (5)
#define LU (6)
#define RU (7)
#define LO (8)
#define RO (9)
#define LM (10)
#define RM (11)
#define MO (12)
#define MU (13)
#define KR (14)
#define KRUO (15)
#define KROU (16)
#define KRH (17)
#define KRV (18)
#define REM (19)
#define SPR (19)

class QString {               // neue Stringdefinition, angelehnt an QT

public:
       char *string;

       QString(char *s="") {                  // Standard-Konstruktor
           string=new char[strlen(s)+1];
           strcpy(string,s);
       }

       QString operator=(char *s) {           // Copy-Konstruktor
           string=new char[strlen(s)+1];
           strcpy(string,s);
       }

       QString operator=(QString s) {           // Copy-Konstruktor
           string=new char[strlen(s.string)+1];
           strcpy(string,s.string);
       }

       QString operator+=(char *s) {           // Verknpfungs-Konstruktor
           char *str=new char[strlen(s)+strlen(string)+1];
           strcpy(str,string);
           strcat(str,s);
           string=str;
       }

       QString operator+=(QString s) {           // Verknpfungs-Konstruktor
           char *str=new char[strlen(s.string)+strlen(string)+1];
           strcpy(str,string);
           strcat(str,s.string);
           string=str;
       }

       bool operator==(char *s) {              // Vergleichsoperator
           if (strcmp(string,s)) return (FALSE);
           else return (TRUE);

       }

       bool operator==(QString s) {              // Vergleichsoperator
           if (strcmp(string,s.data())) return (FALSE);
           else return (TRUE);

       }

       bool operator!=(char *s) {              // Vergleichsoperator
           if (strcmp(string,s)) return (TRUE);
           else return (FALSE);

       }

       int length(void) {               // Gibt die Lnge des Strings zurck.
            return (strlen(string));
       }

       char *data(void) {               // Gibt den String als char* zurck.
            return (string);
       }

       int find(char c, int n=0) {      // Findet ein Zeichen c ab der Stelle n.
           int pos=0;
           char *s=string;
           if (n<0 || n>strlen(string)) return (-1);
           while (n && *s) { s++; n--; }
           do {
              if (*s==c) return(pos);
              s++;
              pos++;
           } while(*s);
           return(-1);               // -1 heit nicht gefunden.
       }

       QString left(int n) {      // Gibt einen String mit den n linkesten Zeichen zurck.
           QString str=string;
           char *s=str.string;
           if (n<0 || n>=strlen(string)) return (str);
           while (n && *s) { s++; n--; }
           *s='\0';
           return(str);
       }

       QString mid(int n, int w) {   // Gibt einen String mit w Zeichen ab der Stelle n zurck.
           QString str=string;
           char *s=str.string;
           if (n<0 || w<0) return (str);
           while (n && *s) { s++; n--; }
           str.string=s;
           while (w && *s) { s++; w--; }
           *s='\0';
           return(str);
       }

       QString right(int n) {    // Gibt einen String mit den n rechtesten Zeichen zurck.
           QString str=string;
           char *s=str.string+str.length();
           if (n<0 || n>=strlen(string)) return (str);
           while (n) { s--; n--; }
           str.string=s;
           return(str);
       }

       int contains(char c) {     // Gibt die Anzahl der Zeichen c im String zurck.
           int pos=0;
           char *s=string;
           do {
              if (*s==c) pos++;
              s++;
           } while(*s);
           return(pos);
       }

       void free(void) {       // Entfernt die Steuerzeichen fr Zeilenumbruch aus einem String.
           char *s=string;
           do {
              if (*s=='\r' || *s=='\n') { *s='\0'; break; }  // String beenden, falls gefunden.
              s++;
           } while(*s);
       }

       void toupper(void) {    // ndert alle Kleinbuchstaben in Grobuchstaben.
           char *s=string;
           bool mark=FALSE;
           do {
               if (*s==';') mark=TRUE;        // Bemerkungen werden nicht gendert.
               if (!mark) {
                  if (*s>=97 && *s<=122) *s=(*s)-32;
               }
               s++;
           } while(*s);
       }

       int free2(void) {                 // ndert einen String fr die Weiterverarbeitung.
           bool space=TRUE;              // Markiert eine Leerfolge
           bool mark=FALSE;              // Markiert eine Bemerkung.
           int i=0, col=0;               // col dient der Spaltenbestimmung beim Lschen von vorhergehenden Leerzeichen.
           char *temp=new char[256];     // Neuer String.
           char *t=temp;
           char *s=string;
           do {
              switch (*s) {
                case '\r':
                case '\n': *s='\0';
                           if (!col) i=1;
                           break;       // Entfernt die Steuerzeichen fr Zeilenumbruch.
                case ' ' :
                case '\t': if (!space) {
                             *t=' ';   // Tabs werden in Leerzeichen gendert.
                             t++;
                             col++;
                             space=TRUE;   // Mehrfache Leerzeichen werden unterbunden.
                           }
                           break;
                case ';' : if (col && !mark) {
                              if (*(t-1)==' ') t--;     // Leerzeichen vor einem ; werden gelscht
                           }
                           *t=*s;
                           t++;
                           col++;
                           mark=TRUE;       // Es folgt eine Bemerkung.
                           space=TRUE;      // Folgende Leerzeichen werden unterbunden.
                           break;
                case ',' :
                case ':' : if (col && !mark) {
                              if (*(t-1)==' ') t--;     // Leerzeichen vor einem : und , werden gelscht
                           }
                           if (!mark) space=TRUE;       // In Bemerkungen werden nachfolgende Leerzeichen erhalten.
                           else space=FALSE;
                           *t=*s;
                           t++;
                           col++;
                           break;
                default  : *t=*s;
                           t++;
                           col++;
                           space=FALSE;
                           break;
              }
              s++;         //
           } while(*s);
           if (*(t-1)==' ') t--;       // Leerzeichen an letzter Stelle wird gelscht.
           *t='\0';                    // String wird beendet.
           // printf("|%s|\n", temp);
           string=temp;
           return (i);
       }
};

int lines=0;          // Maximale Anzahl von Zeilen, die bentigt werden.

QString condJump[10]={"JC", "JNC", "JB", "JNB", "JZ", "JBC", "JNZ", "CJNE", "DJNZ", ""};      // Menge der bedingten Sprnge
QString uncondJump[5]={"JMP", "SJMP", "AJMP", "LJMP", ""};              // Menge der unbedingten Sprnge.
QString Calls[4]={"CALL", "LCALL", "ACALL", ""};                        // Menge der Unterprogrammaufrufe.
QString Moves[4]={"MOV", "MOVX", "MOVC", ""};                     // Menge der Speicheroperationen
QString Math[8]={"ADD", "ADDC", "INC", "SUBB", "DEC", "MUL", "DIV", ""};   // Menge der math. Operationen.

QString types[20]={"", " id=\"bl\"", " id=\"qu\"", " id=\"cl\"", " id=\"hc\"", " id=\"vc\"", " id=\"lu\"", " id=\"ru\"", " id=\"lo\"", " id=\"ro\"", " id=\"lm\"", " id=\"rm\"", " id=\"mo\"", " id=\"mu\"", " id=\"kr\"", " id=\"kruo\"", " id=\"krou\"", " id=\"krh\"", " id=\"krv\"", " id=\"rem\""};
                       // Menge der verschiedenen Bilder fr Verzweigungen und Befehle. (Korrespondiert mit de Makros oben.)

class subprog {        // Liste fr die Sprungmarken

public:
	QString name;
	int pos;
	subprog* next;

};

class structur {       // Inhalt und Art jeder Tabellenzelle

public:
	QString text;         // Inhalt der Zelle
	int type;             // Typ der Zelle
};

class array {             // Eine Zeile in der Tabelle.

public:
	structur field[COLS];    // Es gibt COLS Felder pro Zeile.
};

void convert1(QString file) {          // lscht alle Leerzeichen, Leerzeilen und berflssige Bemerkungen

	QString file2="tmpfile0.txt";         // Ausgabedatei

    FILE *fr, *fw;

	if (!(fr=fopen(file.data(),"r"))) {
		puts("Could not open file to read. Press key to end.");
		exit(1);
	}

 	if (!(fw=fopen(file2.data(),"w"))) {
		puts("Could not open file to write. Press key to end.");
		fclose(fr);
		exit(1);
	}

    char *check;
	QString line("                                                                                                                                                                                                                                                                                            ");
         // Zeile zum Einlesen aus Datei.
    int c;
    bool start=TRUE;     // Flag, da Bemerkungen am Dateianfang nicht ignoriert werden.

    do {
        c=0;
        check=fgets(line.data(),256,fr);          // Zeile auslesen
        c=(line.free2());                         // Zeilen von unntigen Sachen befreien.
        // printf("|%s|\n",line.data());
        if (line.find(';')==0) {
           if (!start) c=-1;            // Zeilen-Bemerkungen in Dateimitte werden ignoriert.
        }
        else start=FALSE;               // Folgende Zeilen-Bemerkungen werden ignoriert.
        if (!check) c=-1;               // Dateiende.
        if (!c) {
           line.toupper();              // Zeile in Grobuchstaben umformen.
           // printf("|%s|\n",line.data());
           fputs(line.data(),fw);       // Zeile schreiben
           fputs("\r\n",fw);            // Zeilenumbruch schreiben
        }
	} while (check);

	fclose(fw);
	fclose(fr);

}

void convert2(void) {                   // fgt an der richtigen Stelle Zeilenumbrche ein.
	int z, col=0;
	bool mark=FALSE;                    // Flag fr Bemerkungen.

    QString file="tmpfile0.txt";        // Eingabedatei
	QString file2="tmpfile.txt";        // Ausgabedatei, Vorlage fr das HTML-File

    FILE *fr, *fw;

	if (!(fr=fopen(file.data(),"r"))) {
		puts("Could not open file to read. Press key to end.");
		exit(1);
	}

 	if (!(fw=fopen(file2.data(),"w"))) {
		puts("Could not open file to write. Press key to end.");
		fclose(fr);
		exit(1);
	}

	do {
		z=getc(fr);
		// printf("%c",z);

		switch (z) {
           case ';' :  if (col && !mark) {      // Semikolons in Bemerkungen ignorieren
                          fputs("\r\n;",fw);    // Bemerkungen alleine in eine Zeile setzen.
                          col=0;
                       }
                       else putc(z,fw);
                       mark=TRUE;
                       col++;
                       break;
			case '\r': break;
			case '\n': if (col) {
                          fputs("\r\n",fw);     // Zeilenumbrche nur einfgen, wenn nicht schon einer vorher war.
                          col=0;
			              lines++;
                          mark=FALSE;
                       }
                       break;
			case ':':  if (!mark) {           // Doppelpunkte in Bemerkungen ignorieren.
                         fputs(":\r\n",fw);     // Sprungmarken allein in eine Zeile schreiben.
                         mark=FALSE;
                         col=0;
                       }
                       else {
                          putc(z,fw);
                          col++;
                       }
					   break;
			case EOF:  break;
			default:   putc(z,fw);
                       col++;
                       break;
		}
	} while (z!=EOF);

	fclose(fw);
	fclose(fr);
}

bool element(QString e, QString *m) {           // gibt an, ob ein Befehl in einer der Menge oben vorkommt.
	do {
		if (e==m->data()) return (TRUE);     // Befehl kommt in der Menge vor.
		m++;
	} while (*m!="");      // "" markiert Mengenende.
	return (FALSE);
}

int element(QString e, subprog *m) {              // gibt Position einer Sprungmarke in der Liste aller Sprungmarken zurck
	// printf("|%s|\n",e.data());
	while (m) {
		// printf("|%s|\n",(m->name).data());
		if (e==m->name) return(m->pos);
		m=m->next;
	}
	return (0);      // Sprungmarke ist nicht in der Liste gespeichert.
}

/* Nachfolgende Prozedur bildet aus einem bereits existierendem Feld des Typs i und einem
berschreibenden Feld des Typs j einen neuen Typ erg, je nachdem, welche Verzweigungen
verknpft werden. Die Variable oben ist dafr da, um zu erkennen, von wo aus ein
Feldverknpfung gelscht werden soll. Siehe dazu auch die Datei changes.html. */

int change(int i, int j, bool oben=FALSE) {
	int res=19*i+j, erg=0;       // res errechnet eine eindeutige Zahl, welcher Felder verknpft werden sollen.
    if (oben && j==EM) {         // Fr Fleder die von "oben" aus gelscht werden sollen, gelten andere Regeln.
      // puts("*");
      switch (res) {
		case 95: 		erg=EM; break;
		case 190:		erg=LO; break;
		case 209:		erg=RO; break;
		case 228:
		case 266:       erg=MO; break;
        case 247:
		case 323:
		case 342:       erg=HC; break;
        default: 		erg=i; break;
      }
    }
    else {
	  switch (res) {
		case 95: 		erg=EM; break;
		case 190:		erg=LU; break;
		case 209:		erg=RU; break;
		case 247:
		case 266:       erg=MU; break;
        case 228:
		case 323:
		case 342:       erg=HC; break;
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
		case 6:
		case 7:
		case 8:
		case 9:
		case 10:
		case 11:
		case 12:
		case 13:
		case 14:
		case 15:
		case 16:
		case 17:
		case 18:		erg=j; break;
		case 99: 		erg=KRV; break;
		case 81: 		erg=KRH; break;
		case 179:
		case 175:
		case 161:
		case 156:
		case 84:
		case 85: 		erg=MO; break;
		case 139:
		case 137:
		case 118:
		case 121:
		case 82:
		case 83:		erg=MU; break;
		case 178:
		case 176:
		case 142:
		case 138:
		case 102:
		case 104:		erg=RM; break;
		case 158:
		case 157:
		case 122:
		case 119:
		case 101:
		case 103:		erg=LM; break;
		case 177:
		case 123:		erg=KRUO; break;
		case 159:
		case 141:		erg=KROU; break;
		case 233:
		case 234:
		case 235:
		case 252:
		case 255:
		case 256:
		case 197:
		case 199:
		case 215:
		case 217:
		case 194:
		case 213:		erg=KR; break;
		default: 		erg=i; break;
	  }
    }
	// printf("%d %d %d %d\n",i,j,res,erg);
	return(erg);     // Neuen Feldtyp zurckgeben.
}

void build(void) {                 // Haupteil, erstellt aus der umgeformten Assembler-Datei die HTML-Tabelle.
	QString file="tmpfile.txt";          // Eingabedatei
	QString file2="structemp.htm";       // Ausgabedatei (unoptimiert)
	QString line("                                                                                                                                                                                                                                                                                            ");
	QString tempStr;
	QString rel, adr, adr2;       // Angabe von Befehlen, relativen Adressen, Werten und Sprungmarken

    int out1=2, out2=1;        // Position, wo rechts (out2) und links (out1) Verzweigungen gelegt werden.

	int j=0, k=0, l=2;

	array row[4*lines];           // Reserviert ausreichend Zeilen fr die Tabelle.

	for (j=0;j<4*lines;j++)
		for (k=0;k<COLS;k++) {
			row[j].field[k].text="";       // Alle Felder ohne Text.
			row[j].field[k].type=EM;       // Alle Felder ohne Typ.
		}

    FILE *fr, *fw;

	if (!(fr=fopen(file.data(),"r"))) {
		puts("Could not open file to read. Press key to end.");
		exit(1);
	}

	char *check;

	subprog *s1=(subprog*)malloc(sizeof(subprog)); // Anfang setzen fr Liste der Sprungmarken
	s1->name="";
	s1->next=0;
	s1->pos=0;
	subprog *anc=s1;       // Anker setzen auf Liste der Sprungmarken

    /* Nachfolgende Routine erstellt erst einmal ein grobes Bild des Struktogramms.
    Dabei werden nur die richtigen Bilder fr Befehle, Sprungmarken und Bemerkungen
    eingefgt. Verzweigungen und genaue Befehlsbeschreibung erfolgt spter.
    Das Ergebnis wird in die Datei tempfile.htm bertragen. */

	do {
        check=fgets(line.data(),256,fr);    // Zeile aus Datei lesen
        line.free();                        // Zeile von Zeilenumbruch befreien.
        // printf("|%s| \n",line.data());
        if (check) {
			if (line.find(';')+1) {         // Zeile enthlt eine Bemerkungen
                line=line.right(line.length()-1);        // Bemerkung ohne fhrendes Semikolon bernehmen.
                if (row[l].field[SPRUNG].type==SPR) {    // Bemerkungen nach Sprungmarken auf gleicher Hhe einfgen
                   row[l].field[REMARKS].type=REM;       // Bemerkungsfeld erzeugen
                   row[l].field[REMARKS].text+=line;     // Bemerkungen hinzufgen
                   row[l].field[REMARKS].text+="<br>";   // Absatz hinzufgen
                }
                else {                                   // Bemerkung in Hhe des vorhergehenden Befehls setzen.
                   // printf("|%s| %d\n",line.data(),line.length());
                   row[l-2].field[REMARKS].type=REM;        // Bemerkungsfeld erzeugen
                   row[l-2].field[REMARKS].text+=line;      // Bemerkungen hinzufgen
                   row[l-2].field[REMARKS].text+="<br>";    // Absatz hinzufgen
                   // printf("%X \n",&(row[l-2].field[REMARKS].text));
                }
            }
			else if (line.find(':')+1) {          // Zeile enthlt eine Sprungmarke
                row[l].field[SPRUNG].type=SPR;             // Sprungmarkenfeld erzeugen
				row[l].field[SPRUNG].text+="<u>";
                row[l].field[SPRUNG].text+=line;           // Sprungemarke unterstrichen hinzufgen
                row[l].field[SPRUNG].text+="</u><br>";
				s1->name=line.left(line.find(':'));        // Sprungmarke ohen Doppelpunkt in die Sprungmarkenliste hinzufgen
				s1->pos=l;                                 // Position der Sprungmarke in der Tabelle merken
				subprog *tmpS=(subprog*)malloc(sizeof(subprog));
				s1->next=tmpS;
				s1=tmpS;
				s1->name="";
				s1->next=0;
				s1->pos=0;
			}
			else {
		      if (line.contains(' ')) tempStr=line.left(line.find(' '));        // Assembler-Befehl herausfiltern, falls mglich.
		      else tempStr=line;
			  if (element(tempStr,condJump)) {      // Befehl war ein bedingter Sprung
					if (tempStr=="DJNZ") {            // Befehl ist DJNZ. Sonderfall, da hier zwei Dinge auf einmal geschehen.
						row[l].field[MID].type=VC;    // Verbindungslinie setzen
						l++;
						row[l].field[MID].type=VC;    // Verbindungslinie setzen
						l++;
					}
					row[l].field[MID].type=QU;        // Fragefeld (spitz) erzeugen.
					row[l].field[MID].text=line;      // gesamte Zeile eintragen.
					l++;
					row[l].field[MID].type=VC;       // Verbindungslinie setzen
					l++;
              }
			  else if (element(tempStr,Calls) || tempStr=="RET"|| tempStr=="RETI") {   // Befehl war ein Unterprogrammaufruf oder ein Rcksprung.
					row[l].field[MID].type=CL;        // Unterprogrammfeld (rund) erzeugen.
					row[l].field[MID].text=line;      // gesamte Zeile eintragen.
					l++;
					row[l].field[MID].type=VC;     // Verbindungslinie setzen
					l++;
			  }
			  else {                            // Befehl war etwas anderes
					row[l].field[MID].type=BL;        // Befehlsfeld (eckig) erzeugen.
					row[l].field[MID].text=line;      // gesamte Zeile eintragen.
					l++;
					row[l].field[MID].type=VC;      // Verbindungslinie setzen
					l++;
			  }
			}
		}
	} while (check);
	fclose(fr);

 	if (!(fw=fopen("tempfile.htm","w"))) {
		puts("Could not open file to write. Press key to end.");
		fclose(fr);
		exit(1);
	}

	// Standard HTML Begin

	tempStr="<html><head>\r\n<title>Struktogramm</title>\r\n<link rel=\"stylesheet\" ";
	tempStr+="type=\"text/css\" href=\"struc.css\">\r\n</head>\r\n";
	tempStr+="<body bgcolor=\"#FFFFFF\">\r\n<table cellpadding=\"0\" cellspacing=\"0\" align=\"center\">\r\n<tr>\r\n";
	// tempStr+="<td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td>\r\n</tr><tr>\r\n";

	fputs(tempStr.data(),fw);

	// Write to HTML

	for (l=0;l<4*lines;l++) {
        // printf("|%s|\n",(row[l].field[MID].text).data());
		for (j=0;j<COLS;j++) {
			tempStr="<td";
			tempStr+=types[row[l].field[j].type];
			tempStr+=">";
			tempStr+=row[l].field[j].text;
			tempStr+="</td>\r\n";
			fputs(tempStr.data(),fw);
		}
		tempStr="</tr><tr>\r\n";
		fputs(tempStr.data(),fw);
	}

	// Standard HTML END

	tempStr="</tr>\r\n</table>\r\n</body>\r\n</html>\r\n";
	fputs(tempStr.data(),fw);

    fclose(fw);

    /* Nachfolgende Routine wandelt die grob erzeugte Tabelle in eine feine mit Verzweigungen um. */

	for (l=2;l<4*lines;l=l+2) {            // Jede zweite Zeile in der Tabelle durchlaufen.
                                           // Die ungeraden Zeilen enthalten nur Verbindungen und keine Befehle.
		tempStr=row[l].field[MID].text;       // Zelleninhalt auslesen
        // printf("|%s|\n",tempStr.data());
		if (tempStr.contains(' ')) tempStr=tempStr.left(tempStr.find(' '));   // Assemblerbefehl herausfiltern, falls mglich.
		if (tempStr!="") {               // Zelleninhalt war nicht leer.
			// printf("|%s|\n",tempStr.data());
			if (element(tempStr,Moves)) {       // Befehl war eine Speicheroperation der Art "tempStr adr, rel"
					rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
					adr=rel.left(rel.find(','));
					rel=rel.right(rel.length()-adr.length()-1);
					row[l].field[MID].text=adr;
                    if (adr=="A") row[l].field[MID].text+="CC";       // Akku als ACC darstellen
                    else if (adr=="C") row[l].field[MID].text+="ARRY";   // Carrybit als CARRY darstellen
					row[l].field[MID].text+=" = ";
					row[l].field[MID].text+=rel;
                    if (rel=="A") row[l].field[MID].text+="CC";       // Akku als ACC darstellen
                    else if (rel=="C") row[l].field[MID].text+="ARRY";    // Carrybit als CARRY darstellen
                    if (tempStr=="MOVX") row[l].field[MID].text+="<BR>(Externer Datenspeicher)";
                    else if (tempStr=="MOVC") row[l].field[MID].text+="<BR>(Programmspeicher)";
            }
			else if (element(tempStr,Calls)) {   // Befehl war ein Unterprogrammaufruf der Art "tempstr rel"
					rel=(row[l].field[MID].text).right((row[l].field[MID].text).length()-tempStr.length()-1);
					row[l].field[MID].text="<u>";
					row[l].field[MID].text+=rel;       // Sprungmarke unterstrichen als Aufruf eintragen
					row[l].field[MID].text+="</u>";
			}
			else if (element(tempStr,uncondJump)) {   // Befehl war ein unbedingter Sprung der Art "tempstr rel"
				rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
				j=element(rel, anc);   // Position der Sprungmarke, zu der gesprungen wird, herausfinden.
				if (j) {
                    row[l-1].field[MID].text+="<br><img src=\"em.gif\" height=\"1\" width=\"85\" border=\"0\">";
                    row[l-1].field[MID].text+=rel;     // Sprungmarke an Verzweigung schreiben.
					if (j>l) {        // Sprungmarkenposition liegt hinter der aktuellen Position => nach unten verzweigen
						row[l].field[MID].type=EM;
						row[l].field[MID].text="";
						row[l+1].field[MID].type=change(row[l+1].field[MID].type,EM,TRUE);
					    row[l-1].field[MID].type=change(row[l-1].field[MID].type,EM);
					    row[l-1].field[MID].type=change(row[l-1].field[MID].type,LU);
					    for (k=1;k<out2;k++) row[l-1].field[MID+k].type=change(row[l-1].field[MID+k].type,HC);
					    row[l-1].field[MID+out2].type=change(row[l-1].field[MID+out2].type,RO);
						for (k=l;k<=j-2;k++) row[k].field[MID+out2].type=change(row[k].field[MID+out2].type,VC);
						row[j-1].field[MID].type=change(row[j-1].field[MID].type,LO);
						for (k=1;k<out2;k++) row[j-1].field[MID+k].type=change(row[j-1].field[MID+k].type,HC);
						row[j-1].field[MID+out2].type=change(row[j-1].field[MID+out2].type,RU);
					}
					else if (j<l) {        // Sprungmarkenposition liegt vor der aktuellen Position => nach oben verzweigen
						row[l].field[MID].type=EM;
						row[l].field[MID].text="";
						row[l+1].field[MID].type=change(row[l+1].field[MID].type,EM,TRUE);
					    row[l-1].field[MID].type=change(row[l-1].field[MID].type,EM);
					    row[l-1].field[MID].type=change(row[l-1].field[MID].type,LU);
					    for (k=1;k<out2;k++) row[l-1].field[MID+k].type=change(row[l-1].field[MID+k].type,HC);
					    row[l-1].field[MID+out2].type=change(row[l-1].field[MID+out2].type,RU);
						for (k=l-2;k>=j;k--) row[k].field[MID+out2].type=change(row[k].field[MID+out2].type,VC);
						row[j-1].field[MID].type=change(row[j-1].field[MID].type,LO);
						for (k=1;k<out2;k++) row[j-1].field[MID+k].type=change(row[j-1].field[MID+k].type,HC);
						row[j-1].field[MID+out2].type=change(row[j-1].field[MID+out2].type,RO);
					}
					else {	// j==l  => Sprung auf sich selbst (Endlosschleife!)
						row[l].field[MID].text="";
						row[l].field[MID].type=RM;
						row[l].field[MID-1].type=LO;
						row[l+1].field[MID].type=change(row[l+1].field[MID].type,EM,TRUE);
						row[l+1].field[MID].type=change(row[l+1].field[MID].type,RU);
						row[l+1].field[MID-1].type=LU;
					}
                    out2=out2%(MID-2)+1;   // Nchstes freies Feld auf der rechten Seite fr Verzweigungen bestimmen.
				}
			}
			else if (element(tempStr,condJump)) {   // Befehl war ein bedingter Sprung (if-Abfragen sind selbsterklrend).
				if (tempStr=="JC") {
					rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
					adr="";
					row[l].field[MID].text="CARRY gesetzt ?";
				}
				if (tempStr=="JNC") {
					rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
					adr="";
					row[l].field[MID].text="CARRY nicht gesetzt ?";
				}
				if (tempStr=="JZ") {
					rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
					adr="";
					row[l].field[MID].text="ACC = #0 ?";
				}
				if (tempStr=="JNZ") {
					rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
					adr="";
					row[l].field[MID].text="ACC != #0 ?";
				}
				if (tempStr=="JB") {
					rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
					adr=rel.left(rel.find(','));
					rel=rel.right(rel.length()-adr.length()-1);
					row[l].field[MID].text=adr;
					row[l].field[MID].text+=" gesetzt ?";
				}
				if (tempStr=="JNB") {
					rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
					adr=rel.left(rel.find(','));
					rel=rel.right(rel.length()-adr.length()-1);
					row[l].field[MID].text=adr;
					row[l].field[MID].text+=" nicht gesetzt ?";
				}
				if (tempStr=="JBC") {
					rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
					adr=rel.left(rel.find(','));
					rel=rel.right(rel.length()-adr.length()-1);
					row[l].field[MID].text=adr;
					row[l].field[MID].text+=" gesetzt ?";
					row[l].field[MID-out1].text="loesche ";
					row[l].field[MID-out1].text+=adr;
					row[l].field[MID-out1].type=BL;
				}
				if (tempStr=="CJNE") {
					rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
					adr=rel.left(rel.find(','));
					adr2=rel.mid(adr.length()+1,rel.find(',', adr.length()+1));
					rel=rel.right(rel.length()-adr.length()-adr2.length()-2);
					row[l].field[MID].text=adr;
					if (adr=="A") row[l].field[MID].text+="CC";      // Akku als ACC darstellen
					row[l].field[MID].text+=" != ";
					row[l].field[MID].text+=adr2;
					row[l].field[MID].text+=" ?";

				}
				if (tempStr=="DJNZ") {
					rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
					adr=rel.left(rel.find(','));
					rel=rel.right(rel.length()-adr.length()-1);
					row[l-2].field[MID].text=adr;
					row[l-2].field[MID].text+=" = ";
					row[l-2].field[MID].text+=adr;
					row[l-2].field[MID].text+=" - #1";
					row[l-2].field[MID].type=BL;
					row[l].field[MID].text=adr;
					row[l].field[MID].text+=" != #0 ?";
				}
				// printf("|%s| |%s| |%s|  ",tempStr.data(), adr.data(), rel.data());
				j=element(rel, anc);      // Position der Sprungmarke, zu der gesprungen wird, herausfinden.
                // printf("%d\n",j);
				if (j) {
					row[l+1].field[MID].text="<img src=\"em.gif\" height=\"1\" width=\"40\" border=\"0\">nein<br><img src=\"em.gif\" height=\"8\" width=\"1\" border=\"0\">";
					row[l].field[MID-1].text="<img src=\"em.gif\" height=\"1\" width=\"100\" border=\"0\">ja<br><img src=\"em.gif\" height=\"1\" width=\"85\" border=\"0\">";
					row[l].field[MID-1].text+=rel;    // Sprungmarke an Verzweigung schreiben.
					for (k=1;k<out1;k++) row[l].field[MID-k].type=change(row[l].field[MID-k].type,HC);
					if (j>l) {       // Sprungmarkenposition liegt hinter der aktuellen Position => nach unten verzweigen
                        row[l].field[MID-out1].type=change(row[l].field[MID-out1].type,LO);
						for (k=l+1;k<=j-2;k++) row[k].field[MID-out1].type=change(row[k].field[MID-out1].type,VC);
						row[j-1].field[MID-out1].type=change(row[j-1].field[MID-out1].type,LU);
						for (k=1;k<out1;k++) row[j-1].field[MID-k].type=change(row[j-1].field[MID-k].type,HC);
						row[j-1].field[MID].type=change(row[j-1].field[MID].type,RO);
					}
					else if (j<=l) {    // Sprungmarkenposition liegt vor der aktuellen Position => nach oben verzweigen
                        row[l].field[MID-out1].type=change(row[l].field[MID-out1].type,LU);
						for (k=l-1;k>=j;k--) row[k].field[MID-out1].type=change(row[k].field[MID-out1].type,VC);
						row[j-1].field[MID-out1].type=change(row[j-1].field[MID-out1].type,LO);
						for (k=1;k<out1;k++) row[j-1].field[MID-k].type=change(row[j-1].field[MID-k].type,HC);
						row[j-1].field[MID].type=change(row[j-1].field[MID].type,RO);
					}
                    out1--;
                    out1=out1%(MID-2)+2;   // Nchstes freies Feld auf der linken Seite fr Verzweigungen bestimmen.
				}
			}
			else if (element(tempStr,Math)) {   // Befehl war eine math. Operation
                if (tempStr=="INC" || tempStr=="DEC" || tempStr=="MUL" || tempStr=="DIV") {   // Ein-Argument-Befehle
                   rel=(row[l].field[MID].text).right((row[l].field[MID].text).length()-tempStr.length()-1);
                   if (tempStr=="INC" || tempStr=="DEC") {
                      row[l].field[MID].text=rel;
                      if (rel=="A") row[l].field[MID].text+="CC";      // Akku als ACC darstellen
                      row[l].field[MID].text+=" = ";
                      row[l].field[MID].text+=rel;
                      if (rel=="A") row[l].field[MID].text+="CC";      // Akku als ACC darstellen
                      if (tempStr=="INC") row[l].field[MID].text+=" + #1";
                      else row[l].field[MID].text+=" - #1";
                   }
                   else {
                        row[l].field[MID].text=tempStr;
                        row[l].field[MID].text+=rel;
                   }
                }
                else {    // Zwei-Argument-Befehle
                    rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
                    adr=rel.left(rel.find(','));
                    rel=rel.right(rel.length()-adr.length()-1);
                    if (tempStr=="ADD" || tempStr=="ADDC") {
                      row[l].field[MID].text=adr;
                      if (adr=="A") row[l].field[MID].text+="CC";      // Akku als ACC darstellen
                      row[l].field[MID].text+=" = ";
                      row[l].field[MID].text+=adr;
                      if (adr=="A") row[l].field[MID].text+="CC";      // Akku als ACC darstellen
                      row[l].field[MID].text+=" + ";
                      row[l].field[MID].text+=rel;
                      if (tempStr=="ADDC") row[l].field[MID].text+=" + CARRY";
                    }
                    else if (tempStr=="SUBB") {
                      row[l].field[MID].text=adr;
                      if (adr=="A") row[l].field[MID].text+="CC";      // Akku als ACC darstellen
                      row[l].field[MID].text+=" = ";
                      row[l].field[MID].text+=adr;
                      if (adr=="A") row[l].field[MID].text+="CC";      // Akku als ACC darstellen
                      row[l].field[MID].text+=" - ";
                      row[l].field[MID].text+=rel;
                      // row[l].field[MID].text+=" - CARRY";
                    }
                }
            }
			else if (tempStr=="ORG") {
                rel=(row[l].field[MID].text).right((row[l].field[MID].text).length()-tempStr.length()-1);
                row[l].field[SPRUNG].type=SPR;
				row[l].field[SPRUNG].text+="(ADR: ";
                row[l].field[SPRUNG].text+=rel;
                row[l].field[SPRUNG].text+=")";
				row[l].field[MID].text="";
				row[l].field[MID].type=EM;
				row[l+1].field[MID].type=change(row[l+1].field[MID].type,EM,TRUE);
            }
			else if (tempStr=="SWAP") {
                rel=(row[l].field[MID].text).right((row[l].field[MID].text).length()-tempStr.length()-1);
				row[l].field[MID].text="Tausche Low und High<BR>Nibble von ";
                row[l].field[MID].text+=rel;
                if (rel=="A") row[l].field[MID].text+="CC";      // Akku als ACC darstellen
            }
			else if (tempStr=="XCH" || tempStr=="XCHD") {
			   rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
			   adr=rel.left(rel.find(','));
			   rel=rel.right(rel.length()-adr.length()-1);
               if (tempStr=="XCHD") {
			            row[l].field[MID].text="tausche Low Nibble von ";
			            row[l].field[MID].text+=adr;
			            if (adr=="A") row[l].field[MID].text+="CC";      // Akku als ACC darstellen
			            row[l].field[MID].text+=" und ";
			            row[l].field[MID].text+=rel;
               }
               else {
			            row[l].field[MID].text="tausche ";
			            row[l].field[MID].text+=adr;
			            if (adr=="A") row[l].field[MID].text+="CC";      // Akku als ACC darstellen
			            row[l].field[MID].text+=" und ";
			            row[l].field[MID].text+=rel;
               }
			}
			else if (tempStr=="NOP") {
				row[l].field[MID].text="Leerbefehl";
			}
			else if (tempStr=="END") {       // Bei Ende wird vorhergehende Verzweigung gelscht.
				row[l].field[MID].text="";
				row[l].field[MID].type=EM;
				row[l+1].field[MID].type=EM;
			}
			else if (tempStr=="RET" || tempStr=="RETI") {
				row[l].field[MID].text="Return";
                if (tempStr=="RETI") row[l].field[MID].text+=" from Interrupt";
				row[l+1].field[MID].type=change(row[l+1].field[MID].type,EM,TRUE);
			}
			else if (tempStr=="SETB") {
                rel=(row[l].field[MID].text).right((row[l].field[MID].text).length()-tempStr.length()-1);
				row[l].field[MID].text="setze ";
				if (rel=="C") row[l].field[MID].text+=" CARRY";   // Carrybit als CARRY darstellen
                else row[l].field[MID].text+=rel;
			}
			else if (tempStr=="CLR") {
                rel=(row[l].field[MID].text).right((row[l].field[MID].text).length()-tempStr.length()-1);
				row[l].field[MID].text="loesche ";
				if (rel=="C") row[l].field[MID].text+=" CARRY";   // Carrybit als CARRY darstellen
                else if (rel=="A") row[l].field[MID].text+="ACC";      // Akku als ACC darstellen
                else row[l].field[MID].text+=rel;
			}
			else if (tempStr=="CPL") {
                rel=(row[l].field[MID].text).right((row[l].field[MID].text).length()-tempStr.length()-1);
				row[l].field[MID].text="invertiere ";
				if (rel=="C") row[l].field[MID].text+="CARRY";
                else if (rel=="A") row[l].field[MID].text+="ACC";      // Akku als ACC darstellen
                else row[l].field[MID].text+=rel;
			}
            else {       // Unbekannter Befehl oder Aliaszuweisung
                if ((row[l].field[MID].text).contains(' ')==2) {
                   rel=(row[l].field[MID].text).mid(tempStr.length()+1, (row[l].field[MID].text).length()-tempStr.length()-1);
                   adr=rel.left(rel.find(' '));
                   rel=rel.right(rel.length()-adr.length()-1);
                   // printf("%s %s %s\n",tempStr, adr, rel);
                   if (adr=="DATA" || adr=="BIT" || adr=="EQU") {
                      row[l].field[MID].text=adr;
                      row[l].field[MID].text+=" ";
                      row[l].field[MID].text+=tempStr;
                      row[l].field[MID].text+=" := ";
                      row[l].field[MID].text+=rel;
                   }
                }
            }
		}
	}

 	if (!(fw=fopen(file2.data(),"w"))) {
		puts("Could not open file to write. Press key to end.");
		fclose(fr);
		exit(1);
	}

	// Standard HTML Begin

	tempStr="<html><head>\r\n<title>Struktogramm</title>\r\n<link rel=\"stylesheet\" ";
	tempStr+="type=\"text/css\" href=\"struc.css\">\r\n</head>\r\n";
	tempStr+="<body bgcolor=\"#FFFFFF\">\r\n<table cellpadding=\"0\" cellspacing=\"0\" align=\"center\" border=\"0\">\r\n<tr>\r\n";
	// tempStr+="<td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td>\r\n</tr><tr>\r\n";

	fputs(tempStr.data(),fw);

	// Write to HTML

	for (l=0;l<4*lines;l++) {
		for (j=0;j<COLS;j++) {
			tempStr="<td";
			tempStr+=types[row[l].field[j].type];
			tempStr+=">";
			tempStr+=row[l].field[j].text;
			tempStr+="</td>\r\n";
			fputs(tempStr.data(),fw);
		}
		tempStr="</tr><tr>\r\n";
		fputs(tempStr.data(),fw);
	}

	// Standard HTML END

	tempStr="</tr>\r\n</table>\r\n</body>\r\n</html>\r\n";
	fputs(tempStr.data(),fw);

	fclose(fw);
}

void optimize(QString file1, QString file2) {     // optimiert die HTML-Datei
    file2=file2.left(file2.find('.'));
    file2+=".htm";
	char *check;
    int i;
	QString line("                                                                                                                                                                                                                                                                                            ");
    int anz=0;                   // Anzahl der Leerzellen
    bool readtd=FALSE, readtr=FALSE, readtdrow=FALSE;
                      // Flags, ob eine Leerzelle, eine Leerzeile oder eine gesamte Leerzeile gelesen wurde
    FILE *fr, *fw;

	if (!(fr=fopen(file1.data(),"r"))) {
		puts("Could not open file to read. Press key to end.");
		exit(1);
	}

 	if (!(fw=fopen(file2.data(),"w"))) {
		puts("Could not open file to write. Press key to end.");
		fclose(fr);
		exit(1);
	}
	do {
        check=fgets(line.data(),256,fr);
        line.free();                    // Zeile vom Zeilenumbruch befreien.
        if (line=="<td></td>") {        // Leerzelle gefunden.
           if (!readtd) readtd=TRUE;
           anz++;                       // Anzahl der Leerzellen erhhen.
           readtr=FALSE;                // keine Leerzeile
        }
        else {
           if (anz!=COLS) {             // Es wurden nicht COLS Leerzellen gelesen wurden
              for (i=0;i<anz;i++) fputs("<td></td>\r\n",fw);    // Aufgehobene Leerzellen in Datei schreiben.
           }
           else {
              if (!readtdrow) {         // Es gab noch keine gesamte Leerzeile.
                 for (i=0;i<anz;i++) fputs("<td></td>\r\n",fw);    // eine gesamte Leerzeile schreiben
                 readtdrow=TRUE;         // gesamte Leerzeile gelesen
              }
           }
           if (line=="</tr><tr>") {      // Leerzeile gelesen
              if (!readtr) {             // Falls es noch keine Leerzeile gab, wird eine geschrieben.
                 fputs(line.data(),fw);
 	             fputs("\r\n",fw);
                 readtr=TRUE;            // Leerzeile gelesen
              }
           }
           else {
                fputs(line.data(),fw);
 	            fputs("\r\n",fw);
                readtr=FALSE;
                readtdrow=FALSE;
           }
           readtd=FALSE;
           anz=0;
        }
    } while (check);

	fclose(fw);
    fclose(fr);
}


int main(int argc, char **argv) {
    int i;
    QString temp("");

    for (i=1; i<argc; i++) {
        temp=argv[i];
        puts("Konvertiere v.txt in tmpfile0.txt.");
        convert1(temp);
        puts("Konvertiere tmpfile.txt in tmpfile.txt.");
	    convert2();
        puts("Erstelle Struktogramm aus tmpfile.txt in structemp.htm.");
	    build();
        puts("Optimiere structemp.htm zu struc.htm.");
	    optimize("structemp.htm", temp);
        puts("FERTIG: Struktogramm erstellt!\n");
    }
	puts("Fertig!");
    getchar();

	return(0);			// executes the application
}


